# - Message Passing Interface (MPI) module.
# 
# The Message Passing Interface (MPI) is a library used to write
# high-performance parallel applications that use message passing, and
# is typically deployed on a cluster. MPI is a standard interface
# (defined by the MPI forum) for which many implementations are
# available. All of these implementations have somewhat different
# compilation approaches (different include paths, libraries to link
# against, etc.), and this module tries to smooth out those differences.
#
# Unlike FindMPI.cmake, this file will attempt to locate FORTRAN
# bindings for MPI.
#
# This module will set the following variables:
#   FORTRAN_MPI_FOUND                  TRUE if we have found MPI
##   FORTRAN_MPI_COMPILE_FLAGS          Compilation flags for MPI programs
#   FORTRAN_MPI_INCLUDE_PATH           Include path(s) for MPI header
##   FORTRAN_MPI_LINK_FLAGS             Linking flags for MPI programs
##   FORTRAN_MPI_LIBRARY                First MPI library to link against (cached)
##   FORTRAN_MPI_EXTRA_LIBRARY          Extra MPI libraries to link against (cached)
#   FORTRAN_MPI_LIBRARIES              All libraries to link MPI programs against
#

# Try to find the MPI FORTRAN driver program
find_program(FORTRAN_MPI_COMPILER 
  NAMES mpifortran mpif95 mpif90 mpif77
  DOC "MPI compiler. Used only to detect MPI compilation flags.")
mark_as_advanced(FORTRAN_MPI_COMPILER)

if (FORTRAN_MPI_INCLUDE_PATH AND FORTRAN_MPI_LIBRARY)
  # Do nothing: we already have FORTRAN_MPI_INCLUDE_PATH and FORTRAN_MPI_LIBRARY in
  # the cache, and we don't want to override those settings.
elseif (FORTRAN_MPI_COMPILER)
  exec_program(${FORTRAN_MPI_COMPILER} 
    ARGS -showme:compile 
    OUTPUT_VARIABLE FORTRAN_MPI_COMPILE_CMDLINE
    RETURN_VALUE FORTRAN_MPI_COMPILER_RETURN)

  if (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
    # If we appear to have -showme:compile, then we should also have
    # -showme:link. Try it.
    exec_program(${FORTRAN_MPI_COMPILER} 
      ARGS -showme:link
      OUTPUT_VARIABLE FORTRAN_MPI_LINK_CMDLINE
      RETURN_VALUE FORTRAN_MPI_COMPILER_RETURN)

    # Note that we probably have -showme:incdirs and -showme:libdirs
    # as well.
    set(FORTRAN_MPI_COMPILER_MAY_HAVE_INCLIBDIRS TRUE)
  endif (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)

  if (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
    # Do nothing: we have our command lines now
  else (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
    # Older versions of LAM-MPI have "-showme". Try it.
    exec_program(${FORTRAN_MPI_COMPILER} 
      ARGS -showme
      OUTPUT_VARIABLE FORTRAN_MPI_COMPILE_CMDLINE
      RETURN_VALUE FORTRAN_MPI_COMPILER_RETURN)
  endif (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)  

  if (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
    # Do nothing: we have our command lines now
  else (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
    # MPICH uses "-show". Try it.
    exec_program(${FORTRAN_MPI_COMPILER} 
      ARGS -show
      OUTPUT_VARIABLE FORTRAN_MPI_COMPILE_CMDLINE
      RETURN_VALUE FORTRAN_MPI_COMPILER_RETURN)
  endif (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)  

  if (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
    # We have our command lines, but we might need to copy
    # FORTRAN_MPI_COMPILE_CMDLINE into FORTRAN_MPI_LINK_CMDLINE, if the underlying
    if (NOT FORTRAN_MPI_LINK_CMDLINE)
      SET(FORTRAN_MPI_LINK_CMDLINE ${FORTRAN_MPI_COMPILE_CMDLINE})
    endif (NOT FORTRAN_MPI_LINK_CMDLINE)
  else (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
    message(STATUS "Unable to determine MPI from MPI driver ${FORTRAN_MPI_COMPILER}")
  endif (FORTRAN_MPI_COMPILER_RETURN EQUAL 0)
endif (FORTRAN_MPI_INCLUDE_PATH AND FORTRAN_MPI_LIBRARY)

if (FORTRAN_MPI_INCLUDE_PATH AND FORTRAN_MPI_LIBRARY)
  # Do nothing: we already have FORTRAN_MPI_INCLUDE_PATH and FORTRAN_MPI_LIBRARY in
  # the cache, and we don't want to override those settings.
elseif (FORTRAN_MPI_COMPILE_CMDLINE)
  # Extract compile flags from the compile command line.
  string(REGEX MATCHALL "-D([^\" ]+|\"[^\"]+\")" FORTRAN_MPI_ALL_COMPILE_FLAGS "${FORTRAN_MPI_COMPILE_CMDLINE}")
  set(FORTRAN_MPI_COMPILE_FLAGS_WORK)
  foreach(FLAG ${FORTRAN_MPI_ALL_COMPILE_FLAGS})
    if (FORTRAN_MPI_COMPILE_FLAGS_WORK)
      set(FORTRAN_MPI_COMPILE_FLAGS_WORK "${FORTRAN_MPI_COMPILE_FLAGS_WORK} ${FLAG}")
    else(FORTRAN_MPI_COMPILE_FLAGS_WORK)
      set(FORTRAN_MPI_COMPILE_FLAGS_WORK ${FLAG})
    endif(FORTRAN_MPI_COMPILE_FLAGS_WORK)
  endforeach(FLAG)

  # Extract include paths from compile command line
  string(REGEX MATCHALL "-I([^\" ]+|\"[^\"]+\")" FORTRAN_MPI_ALL_INCLUDE_PATHS "${FORTRAN_MPI_COMPILE_CMDLINE}")
  set(FORTRAN_MPI_INCLUDE_PATH_WORK)
  foreach(IPATH ${FORTRAN_MPI_ALL_INCLUDE_PATHS})
    string(REGEX REPLACE "^-I" "" IPATH ${IPATH})
    string(REGEX REPLACE "//" "/" IPATH ${IPATH})
    list(APPEND FORTRAN_MPI_INCLUDE_PATH_WORK ${IPATH})
  endforeach(IPATH)
  
  if (NOT FORTRAN_MPI_INCLUDE_PATH_WORK)
    if (FORTRAN_MPI_COMPILER_MAY_HAVE_INCLIBDIRS)
      # The compile command line didn't have any include paths on it,
      # but we may have -showme:incdirs. Use it.
      exec_program(${FORTRAN_MPI_COMPILER} 
        ARGS -showme:incdirs
        OUTPUT_VARIABLE FORTRAN_MPI_INCLUDE_PATH_WORK
        RETURN_VALUE FORTRAN_MPI_COMPILER_RETURN)
      separate_arguments(FORTRAN_MPI_INCLUDE_PATH_WORK)
    endif (FORTRAN_MPI_COMPILER_MAY_HAVE_INCLIBDIRS)
  endif (NOT FORTRAN_MPI_INCLUDE_PATH_WORK)

  if (NOT FORTRAN_MPI_INCLUDE_PATH_WORK)
    # If all else fails, just search for mpif.h in the normal include
    # paths.
    find_path(FORTRAN_MPI_INCLUDE_PATH mpif.h)
    set(FORTRAN_MPI_INCLUDE_PATH_WORK ${FORTRAN_MPI_INCLUDE_PATH})
  endif (NOT FORTRAN_MPI_INCLUDE_PATH_WORK)

  # Extract linker paths from the link command line
  string(REGEX MATCHALL "-L([^\" ]+|\"[^\"]+\")" FORTRAN_MPI_ALL_LINK_PATHS "${FORTRAN_MPI_LINK_CMDLINE}")
  set(FORTRAN_MPI_LINK_PATH)
  foreach(LPATH ${FORTRAN_MPI_ALL_LINK_PATHS})
    string(REGEX REPLACE "^-L" "" LPATH ${LPATH})
    string(REGEX REPLACE "//" "/" LPATH ${LPATH})
    list(APPEND FORTRAN_MPI_LINK_PATH ${LPATH})
  endforeach(LPATH)

  if (NOT FORTRAN_MPI_LINK_PATH)
    if (FORTRAN_MPI_COMPILER_MAY_HAVE_INCLIBDIRS)
      # The compile command line didn't have any linking paths on it,
      # but we may have -showme:libdirs. Use it.
      exec_program(${FORTRAN_MPI_COMPILER} 
        ARGS -showme:libdirs
        OUTPUT_VARIABLE FORTRAN_MPI_LINK_PATH
        RETURN_VALUE FORTRAN_MPI_COMPILER_RETURN)
      separate_arguments(FORTRAN_MPI_LINK_PATH)
    endif (FORTRAN_MPI_COMPILER_MAY_HAVE_INCLIBDIRS)
  endif (NOT FORTRAN_MPI_LINK_PATH)

  # Extract linker flags from the link command line
  string(REGEX MATCHALL "-Wl,([^\" ]+|\"[^\"]+\")" FORTRAN_MPI_ALL_LINK_FLAGS "${FORTRAN_MPI_LINK_CMDLINE}")
  set(FORTRAN_MPI_LINK_FLAGS_WORK)
  foreach(FLAG ${FORTRAN_MPI_ALL_LINK_FLAGS})
    if (FORTRAN_MPI_LINK_FLAGS_WORK)
      set(FORTRAN_MPI_LINK_FLAGS_WORK "${FORTRAN_MPI_LINK_FLAGS_WORK} ${FLAG}")
    else(FORTRAN_MPI_LINK_FLAGS_WORK)
      set(FORTRAN_MPI_LINK_FLAGS_WORK ${FLAG})
    endif(FORTRAN_MPI_LINK_FLAGS_WORK)
  endforeach(FLAG)

  # Extract the set of libraries to link against from the link command
  # line
  string(REGEX MATCHALL "-l([^\" ]+|\"[^\"]+\")" FORTRAN_MPI_LIBNAMES "${FORTRAN_MPI_LINK_CMDLINE}")

  # Determine full path names for all of the libraries that one needs
  # to link against in an MPI program
  set(FORTRAN_MPI_LIBRARIES)
  foreach(LIB ${FORTRAN_MPI_LIBNAMES})
    string(REGEX REPLACE "^-l" "" LIB ${LIB})
    set(FORTRAN_MPI_LIB "FORTRAN_MPI_LIB-NOTFOUND" CACHE FILEPATH "Cleared" FORCE)
    find_library(FORTRAN_MPI_LIB ${LIB} HINTS ${FORTRAN_MPI_LINK_PATH})
    if (FORTRAN_MPI_LIB)
      list(APPEND FORTRAN_MPI_LIBRARIES ${FORTRAN_MPI_LIB})
    else (FORTRAN_MPI_LIB)
      message(SEND_ERROR "Unable to find MPI library ${LIB}")
    endif (FORTRAN_MPI_LIB)
  endforeach(LIB)
  set(FORTRAN_MPI_LIB "FORTRAN_MPI_LIB-NOTFOUND" CACHE INTERNAL "Scratch variable for MPI detection" FORCE)

  # Chop FORTRAN_MPI_LIBRARIES into the old-style FORTRAN_MPI_LIBRARY and
  # FORTRAN_MPI_EXTRA_LIBRARY.
  list(LENGTH FORTRAN_MPI_LIBRARIES FORTRAN_MPI_NUMLIBS)
  list(LENGTH FORTRAN_MPI_LIBNAMES FORTRAN_MPI_NUMLIBS_EXPECTED)
  if (FORTRAN_MPI_NUMLIBS EQUAL FORTRAN_MPI_NUMLIBS_EXPECTED)
    list(GET FORTRAN_MPI_LIBRARIES 0 FORTRAN_MPI_LIBRARY_WORK)
    set(FORTRAN_MPI_LIBRARY ${FORTRAN_MPI_LIBRARY_WORK} CACHE FILEPATH "MPI library to link against" FORCE)
  else (FORTRAN_MPI_NUMLIBS EQUAL FORTRAN_MPI_NUMLIBS_EXPECTED)
    set(FORTRAN_MPI_LIBRARY "FORTRAN_MPI_LIBRARY-NOTFOUND" CACHE FILEPATH "MPI library to link against" FORCE)
  endif (FORTRAN_MPI_NUMLIBS EQUAL FORTRAN_MPI_NUMLIBS_EXPECTED)
  if (FORTRAN_MPI_NUMLIBS GREATER 1)
    set(FORTRAN_MPI_EXTRA_LIBRARY_WORK ${FORTRAN_MPI_LIBRARIES})
    list(REMOVE_AT FORTRAN_MPI_EXTRA_LIBRARY_WORK 0)
    set(FORTRAN_MPI_EXTRA_LIBRARY ${FORTRAN_MPI_EXTRA_LIBRARY_WORK} CACHE STRING "Extra MPI libraries to link against" FORCE)
  else (FORTRAN_MPI_NUMLIBS GREATER 1)
    set(FORTRAN_MPI_EXTRA_LIBRARY "FORTRAN_MPI_EXTRA_LIBRARY-NOTFOUND" CACHE STRING "Extra MPI libraries to link against" FORCE)
  endif (FORTRAN_MPI_NUMLIBS GREATER 1)

  # Set up all of the appropriate cache entries
  set(FORTRAN_MPI_COMPILE_FLAGS ${FORTRAN_MPI_COMPILE_FLAGS_WORK} CACHE STRING "MPI compilation flags" FORCE)
  set(FORTRAN_MPI_INCLUDE_PATH ${FORTRAN_MPI_INCLUDE_PATH_WORK} CACHE STRING "MPI include path" FORCE)
  set(FORTRAN_MPI_LINK_FLAGS ${FORTRAN_MPI_LINK_FLAGS_WORK} CACHE STRING "MPI linking flags" FORCE)
else (FORTRAN_MPI_COMPILE_CMDLINE)
  find_path(FORTRAN_MPI_INCLUDE_PATH mpif.h 
    /usr/local/include 
    /usr/include 
    /usr/include/mpi
    /usr/local/mpi/include
    "C:/Program Files/MPICH/SDK/Include" 
    "$ENV{SystemDrive}/Program Files/MPICH2/include"
    "$ENV{SystemDrive}/Program Files/Microsoft Compute Cluster Pack/Include"
    )
  
  # Decide between 32-bit and 64-bit libraries for Microsoft's MPI
  if (CMAKE_CL_64)
    set(MS_MPI_ARCH_DIR amd64)
  else (CMAKE_CL_64)
    set(MS_MPI_ARCH_DIR i386)
  endif (CMAKE_CL_64)
  
  find_library(FORTRAN_MPI_LIBRARY 
    NAMES mpi mpich msmpi
    PATHS /usr/lib /usr/local/lib /usr/local/mpi/lib
    "C:/Program Files/MPICH/SDK/Lib" 
    "$ENV{SystemDrive}/Program Files/MPICH/SDK/Lib"
    "$ENV{SystemDrive}/Program Files/Microsoft Compute Cluster Pack/Lib/${MS_MPI_ARCH_DIR}"
    )
  find_library(FORTRAN_MPI_LIBRARY 
    NAMES mpich2
    PATHS
    "$ENV{SystemDrive}/Program Files/MPICH2/Lib")

  find_library(FORTRAN_MPI_EXTRA_LIBRARY 
    NAMES mpi++
    PATHS /usr/lib /usr/local/lib /usr/local/mpi/lib
    "C:/Program Files/MPICH/SDK/Lib" 
    DOC "Extra MPI libraries to link against.")

  set(FORTRAN_MPI_COMPILE_FLAGS "" CACHE STRING "MPI compilation flags")
  set(FORTRAN_MPI_LINK_FLAGS "" CACHE STRING "MPI linking flags")
endif (FORTRAN_MPI_INCLUDE_PATH AND FORTRAN_MPI_LIBRARY)

# on BlueGene/L the MPI lib is named libmpich.rts.a, there also these additional libs are required
if("${FORTRAN_MPI_LIBRARY}" MATCHES "mpich.rts")
   set(FORTRAN_MPI_EXTRA_LIBRARY ${FORTRAN_MPI_EXTRA_LIBRARY} msglayer.rts devices.rts rts.rts devices.rts)
   set(FORTRAN_MPI_LIBRARY ${FORTRAN_MPI_LIBRARY}  msglayer.rts devices.rts rts.rts devices.rts)
endif("${FORTRAN_MPI_LIBRARY}" MATCHES "mpich.rts")

# Set up extra variables to conform to 
if (FORTRAN_MPI_EXTRA_LIBRARY)
  set(FORTRAN_MPI_LIBRARIES ${FORTRAN_MPI_LIBRARY} ${FORTRAN_MPI_EXTRA_LIBRARY})
else (FORTRAN_MPI_EXTRA_LIBRARY)
  set(FORTRAN_MPI_LIBRARIES ${FORTRAN_MPI_LIBRARY})
endif (FORTRAN_MPI_EXTRA_LIBRARY)

if (FORTRAN_MPI_INCLUDE_PATH AND FORTRAN_MPI_LIBRARY)
  set(FORTRAN_MPI_FOUND TRUE)
else (FORTRAN_MPI_INCLUDE_PATH AND FORTRAN_MPI_LIBRARY)
  set(FORTRAN_MPI_FOUND FALSE)
endif (FORTRAN_MPI_INCLUDE_PATH AND FORTRAN_MPI_LIBRARY)

include(FindPackageHandleStandardArgs)
# handle the QUIETLY and REQUIRED arguments 
find_package_handle_standard_args(MPI DEFAULT_MSG FORTRAN_MPI_LIBRARY FORTRAN_MPI_INCLUDE_PATH)

mark_as_advanced(FORTRAN_MPI_INCLUDE_PATH FORTRAN_MPI_COMPILE_FLAGS FORTRAN_MPI_LINK_FLAGS FORTRAN_MPI_LIBRARY 
  FORTRAN_MPI_EXTRA_LIBRARY)
