  ! ------------------------------------------------------------------
  ! PARTICLE STATE
  ! ------------------------------------------------------------------
  module particle_m
  use parallel_m

  ! ------------------------------------------------------------------
  integer num_particles ! number of particles in the simulation
  integer num_neighbors ! number of nearest neighbors to track
  real terminal_force ! the smallest inter-particle force we'll allow.
  real terminal_velocity ! the smallest particle velocity we'll allow.
  integer min_steps ! the smallest number of time steps we'll allow.
  integer simstep  ! the number of steps we've simulated
  real,    allocatable, target, public :: pxyz(:,:) ! particle coordinates
  real,    allocatable, target, public :: pmom(:,:) ! particle momenta
  real,    allocatable, target, public :: pfrc(:,:) ! particle forces
  real,    allocatable, target, public :: mass(:)   ! particle masses
  integer, allocatable, target, public :: nnnn(:,:) ! n nearest neighbor numbers
  real,    allocatable, target, public :: dist(:,:) ! distance to nearest neighbors
  integer, allocatable, target, public :: pown(:)   ! process which owns particle

  ! ------------------------------------------------------------------

  contains

  ! ------------------------------------------------------------------
  ! READ INITIAL PARTICLE STATE
  integer function read_particles( fname )
  implicit none
  character (len=*), intent(in) :: fname
  integer i, j
  integer pfile

  pfile = 1
  open (unit=pfile,file=trim(fname),form='formatted',status='old')
  read(pfile,*) num_particles
  read(pfile,*) num_neighbors
  read(pfile,*) terminal_force
  read(pfile,*) terminal_velocity
  read(pfile,*) min_steps
  allocate( pxyz(3,num_particles) )
  allocate( pmom(3,num_particles) )
  allocate( pfrc(3,num_particles) )
  allocate( mass(num_particles) )
  allocate( nnnn(num_neighbors,num_particles) )
  allocate( dist(num_neighbors,num_particles) )
  allocate( pown(num_particles) )

  !read(pfile,*) ( pown(i), i=1,num_particles )
  read(pfile,*) ( mass(i), i=1,num_particles )
  read(pfile,*) ( ( pxyz(j,i), j=1,3 ), i=1,num_particles )
  read(pfile,*) ( ( pmom(j,i), j=1,3 ), i=1,num_particles )

  do i = 1,num_particles
    do j = 1,3
      pfrc(j,i) = 0.
    enddo
  enddo

1 format( 'Read ', i4, ' particles' )
  if ( local_id .eq. 0 ) write(*,1) num_particles

  read_particles = 0
  end function read_particles
  ! ------------------------------------------------------------------

  ! ------------------------------------------------------------------
  ! DECIDE WHETHER STATE CHANGE IS SIGNIFICANT
  logical function evaluate_exit_condition( dt )

  implicit none
  integer i, j
  real fmag2
  real maxforce2
  real vmag2
  real maxvel2
  real dt

  evaluate_exit_condition = simstep .ge. min_steps
  if ( .not. evaluate_exit_condition ) then
    return
  endif

  maxforce2 = 0.
  maxvel2 = 0.
  do i = 1, num_particles
    fmag2 = 0.
    vmag2 = 0.
    do j = 1, 3
      fmag2 = fmag2 + pfrc(j,i) * pfrc(j,i)
      vmag2 = vmag2 + pmom(j,i) * pmom(j,i)
    end do
    if ( fmag2 .gt. maxforce2 ) then
      maxforce2 = fmag2
    end if
    if ( vmag2 .gt. maxvel2 ) then
      maxvel2 = vmag2
    end if
  end do

  dt = max( 1., 0.1*sqrt(1./maxforce2) )

  evaluate_exit_condition = maxforce2 .gt. terminal_force .and. maxvel2 .gt. terminal_velocity
  end function evaluate_exit_condition

  end module particle_m
  ! ------------------------------------------------------------------
