  ! ------------------------------------------------------------------
  ! PARTICLE RULES: THEIR DEFINITION, STATE, AND APPLICATION
  ! ------------------------------------------------------------------
  module rules_m
  use particle_m
  use geometry_m

  ! ------------------------------------------------------------------
  integer num_edg_springs ! springs connecting 2 particles
  integer num_tri_springs ! springs connecting 3 particles
  integer num_pnt_hinges  ! rotational ('thin'-hinge) springs
  integer num_edg_hinges  ! rotational (piano-hinge) springs
  integer num_fixed_coord ! number of coordinate constraints on particles
  integer num_edge_cstr   ! number of edge constraints (particle must lie on edge) 
  integer num_plane_cstr  ! number of plane constraints (particle must lie on plane) 
  real, allocatable, public :: kedg(:)    ! stiffness of edge-springs
  real, allocatable, public :: ktri(:)    ! stiffness of triangles
  real, allocatable, public :: ltri(:,:)  ! free length of triangle edges
  real, allocatable, public :: jedg(:)    ! stiffness of point-hinges
  real, allocatable, public :: jtri(:)    ! stiffness of piano-hinges
  integer, allocatable, public :: ies(:,:)   ! ids of particles defining edge springs
  integer, allocatable, public :: its(:,:)   ! ids of particles defining tri springs
  integer, allocatable, public :: ieh(:,:)   ! ids of particles defining point hinges
  integer, allocatable, public :: ith(:,:)   ! ids of particles defining edge hinges
  real, allocatable, public :: art(:,:)   ! arrangement of particles wrt tris
  integer, allocatable, public :: tpc(:,:) ! type of particle coordinate constraint
  ! ------------------------------------------------------------------

  contains

  ! ------------------------------------------------------------------
  real function compute_arrangement( pid, tid0, tid1, tid2 )
  implicit none
  integer pid, tid0, tid1, tid2
  real edg0(3)
  real edg1(3)
  real norm(3)
  real peln(3)
  real nmag
  integer i
  nmag = triangle_unit_normal( norm, pxyz(:,tid0), pxyz(:,tid1), pxyz(:,tid2) )
  do i = 1,3
    peln(i) = pxyz(i,pid) - pxyz(i,tid0)
  enddo
  compute_arrangement = 0.
  do i = 1,3
    compute_arrangement = compute_arrangement + norm(i) * peln(i)
  enddo
  end function compute_arrangement
  ! ------------------------------------------------------------------

  ! ------------------------------------------------------------------
  integer function read_rules( fname )
  implicit none
  character (len=*), intent(in) :: fname
  integer rfile
  integer i, j, k
  integer endpt1, endpt2

  rfile = 2
  open( unit=rfile, file=trim(fname), form='formatted', status='old' )
  
  read(rfile,*) num_edg_springs
  read(rfile,*) num_tri_springs
  read(rfile,*) num_pnt_hinges
  read(rfile,*) num_edg_hinges
  read(rfile,*) num_fixed_coord
  read(rfile,*) num_edge_cstr
  read(rfile,*) num_plane_cstr

  allocate( kedg(num_edg_springs) )
  allocate( ktri(num_tri_springs) )
  allocate( ltri(num_tri_springs,3) )
  allocate( jedg(num_pnt_hinges) )
  allocate( jtri(num_edg_hinges) )
  allocate( ies(num_edg_springs,2) )
  allocate( its(num_tri_springs,3) )
  allocate( ieh(num_pnt_hinges,3) )
  allocate( ith(num_edg_hinges,4) )
  allocate( tpc(num_fixed_coord,2) )
  allocate( art(num_particles,num_tri_springs) )

  read(rfile,*) ( kedg(i), i = 1,num_edg_springs )
  read(rfile,*) ( ktri(i), i = 1,num_tri_springs )
  read(rfile,*) ( jedg(i), i = 1,num_pnt_hinges )
  read(rfile,*) ( jtri(i), i = 1,num_edg_hinges )
  read(rfile,*) ( ( ies(i,j), j = 1,2 ), i=1,num_edg_springs )
  read(rfile,*) ( ( its(i,j), j = 1,3 ), i=1,num_tri_springs )
  read(rfile,*) ( ( ieh(i,j), j = 1,3 ), i=1,num_pnt_hinges )
  read(rfile,*) ( ( ith(i,j), j = 1,4 ), i=1,num_edg_hinges )
  read(rfile,*) ( ( tpc(i,j), j = 1,2 ), i=1,num_fixed_coord )

  do i = 1,num_particles
    do j = 1,num_tri_springs
      art(i,j) = compute_arrangement( i, its(j,1), its(j,2), its(j,3) )
    enddo
  enddo
  do i = 1,num_tri_springs
    do j = 1,3
      endpt1 = its(i,j)
      if ( j .lt. 3 ) then
        endpt2 = its(i,j + 1)
      else
        endpt2 = its(i,1)
      endif
      ltri(i,j) = 0.
      do k = 1,3
        ltri(i,j) = ltri(i,j) + ( pxyz(k,endpt2) - pxyz(k,endpt1) )**2
      enddo
      ltri(i,j) = sqrt( ltri(i,j) )
    enddo
    ! write(*,*) 'Triangle ',i, ' conn ', ( its(i,j), j = 1,3 ), ' lengths ', ( ltri(i,j), j = 1,3 )
  enddo

1 format( 'Read ', i4, ' triangles' )
  if ( local_id .eq. 0 ) write(*,1) num_tri_springs

  read_rules = 0
  end function read_rules
  ! ------------------------------------------------------------------

  ! ------------------------------------------------------------------
  ! MOVE PARTICLES BASED ON THEIR CURRENT MOMENTA AND EXTERNAL FORCES
  subroutine advance_particles( dt )
  implicit none
  real, intent(in) :: dt

  integer i, j, k, l
  real nmag
  real norm(3)
  real lng(3)
  real ldir(3,3)
  integer it(3)
  real newmom
  integer pidx
  integer next_pinned_node
  integer next_pinned_coord

  ! Zero out the forces on particles
  do i=1,num_particles
    do j=1,3
      pfrc(j,i) = 0.
    enddo
  enddo
  ! Now sum forces on particles due to springs
  do i = 1,num_tri_springs
    do j = 1,3
      it(j) = its(i,j)
    enddo
    nmag = triangle_unit_normal( norm, pxyz(:,it(1)), pxyz(:,it(2)), pxyz(:,it(3)) )
    do j = 1,3
      lng(j) = 0.
    enddo
    do j = 1,3
      do k = 1,3
        if ( k .lt. 3 ) then
          ldir(k,j) = pxyz(j,it(k+1)) - pxyz(j,it(k))
        else
          ldir(k,j) = pxyz(j,it(1)) - pxyz(j,it(k))
        endif
        lng(k) = lng(k) + ldir(k,j)**2
        ! if ( j .eq. 3 ) write(*,*) ldir(k,3)
      enddo
    enddo
    do j = 1,3
      lng(j) = sqrt( lng(j) )
    enddo
    do j = 1,3
      do k = 1,3
        if ( k .gt. 1 ) then
          pfrc(j,it(k)) = pfrc(j,it(k)) + ktri(i) * &
            ( max( 0., ( lng(k)     - ltri(i,k)     ) ) * ldir(k,j) / lng(k) - &
              max( 0., ( lng(k - 1) - ltri(i,k - 1) ) ) * ldir(k,j) / lng(k - 1) )
          ! if ( j .eq. 3 ) write(*,*) i, k, it(k), &
          !    max( 0., ( lng(k)     - ltri(i,k)     ) ) * ldir(k,j) / lng(k), &
          !   -max( 0., ( lng(k - 1) - ltri(i,k - 1) ) ) * ldir(k,j) / lng(k - 1) 
        else
          pfrc(j,it(k)) = pfrc(j,it(k)) - ktri(i) * &
            ( max( 0., ( lng(k) - ltri(i,k) ) ) * ldir(k,j) / lng(k) - &
              max( 0., ( lng(3) - ltri(i,3) ) ) * ldir(k,j) / lng(3) )
          ! if ( j .eq. 3 ) write(*,*) i, k, it(k), &
          !    max( 0., ( lng(k) - ltri(i,k) ) ) * ldir(k,j) / lng(k), &
          !   -max( 0., ( lng(3) - ltri(i,3) ) ) * ldir(k,j) / lng(3) 
        endif
      enddo
    enddo
  enddo
  ! Add in gravitational force
  do i = 1,num_particles
    pfrc(3,i) = pfrc(3,i) - mass(i) * 0.1 ! 9.81
    ! write(*,*) 'Part ', i, ' force ', (pfrc(j,i), j=1,3)
  enddo
  ! Now integrate the velocity and position, excluding
  ! pinned nodal coordinates as required.
  if ( num_fixed_coord .gt. 0 ) then
    pidx = 1
    next_pinned_node = tpc(pidx,1)
    next_pinned_coord = tpc(pidx,2)
  else
    pidx = 0
  endif
  do i = 1,num_particles
    do j = 1,3
      if ( pidx .ne. 0 .and. ( next_pinned_node .eq. i .and. next_pinned_coord .eq. j ) ) then
        if ( pidx .gt. num_fixed_coord ) then
          pidx = 0
        else
          pidx = pidx + 1
          next_pinned_node = tpc(pidx,1)
          next_pinned_coord = tpc(pidx,2)
        endif
      else
        pxyz(j,i) = pxyz(j,i) + dt * pmom(j,i)
        pmom(j,i) = pmom(j,i) + dt * pfrc(j,i)
      endif
    enddo
  enddo

  end subroutine advance_particles
  ! ------------------------------------------------------------------

  end module rules_m
